﻿// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License. See License.txt in the project root for
// license information.

namespace Microsoft.Azure.Search.Models
{
    using System.Collections.Generic;
    using System.Globalization;
    using System.Linq;
    using Common;
    using Spatial;

    /// <summary>
    /// Represents a parameter value to be used in scoring functions (for example, referencePointParameter).
    /// </summary>
    public class ScoringParameter
    {
        /// <summary>
        /// Initializes a new instance of the ScoringParameter class with the given name and string values.
        /// </summary>
        /// <param name="name">Name of the scoring parameter.</param>
        /// <param name="values">Values of the scoring parameter.</param>
        public ScoringParameter(string name, IEnumerable<string> values)
        {
            Throw.IfArgumentNull(name, "name");
            Throw.IfArgumentNull(values, "values");

            Name = name;
            Values = values.ToList();   // Deep copy.
        }

        /// <summary>
        /// Initializes a new instance of the ScoringParameter class with the given name and GeographyPoint value.
        /// </summary>
        /// <param name="name">Name of the scoring parameter.</param>
        /// <param name="value">Value of the scoring parameter.</param>
        public ScoringParameter(string name, GeographyPoint value) : this(name, ToLonLatStrings(value)) { }

        /// <summary>
        /// Gets the name of the scoring parameter.
        /// </summary>
        public string Name { get; }

        /// <summary>
        /// Gets the values of the scoring parameter.
        /// </summary>
        public IEnumerable<string> Values{ get; }

        /// <summary>
        /// Returns the scoring parameter in a format that can be used in a Search API request.
        /// </summary>
        /// <returns>
        /// The scoring parameter as a colon-separated name-value pair (for example, mylocation:-122.2,44.8)
        /// </returns>
        public override string ToString() =>
            $"{Name}-{Values.Select(v => EscapeValue(v)).ToCommaSeparatedString()}";

        private static string EscapeValue(string value)
        {
            value = value?.Replace("'", "''");
            return $"'{value}'";
        }

        private static IEnumerable<string> ToLonLatStrings(GeographyPoint point)
        {
            if (point == null)
            {
                yield break;
            }

            yield return point.Longitude.ToString(CultureInfo.InvariantCulture);
            yield return point.Latitude.ToString(CultureInfo.InvariantCulture);
        }
    }
}
